#!/usr/bin/python

import sys, numpy, random, copy, time, argparse

# Monte-Carlo simulations of Hex.

# the standard board structure:
#    boards are n x n lists of lists, with each entry being '1'st player or '2'nd player
#    1st player connects horizontally, 2nd player vertically

# the board convention

#    22222
#  1 o-o-o 1
#  1 |/|/| 1
#  1 o-o-o 1
#  1 |/|/| 1
#  1 o-o-o 1
#    22222

#  the indexing convention

#   ----------> 
# | 0,0 0,1 0,2
# | 1,0 1,1 1,2
# v 2,0 2,1 2,2

# example boards

example_first_a  = [[1,1,1],
                    [1,1,1],
                    [1,1,1]];

example_second_a  = [[2,2,2],
                     [2,2,2],
                     [2,2,2]];
 

parser = argparse.ArgumentParser(
            description='Perform Monte-Carlo simulations on Hex.')

parser.add_argument('-size', default=5, metavar='k',type=int,
       help='half-board size for nxn where n=2k+1')

parser.add_argument('-trials', default=1000, metavar='N', type=int,
       help='number of trials to perform for each cell')

parser.add_argument('-sampling', default='hex', choices=['bernoulli','hex'],
       help="which sampling method to use")
       
       
args = parser.parse_args()


def extend(board):
    sideLength = len(board);

    extended_board = [];

    first_row = [2] + ( [1] * (sideLength + 1) )
    extended_board.append(first_row);

    for row in board:
        middle_row = [2] + row + [2];
        extended_board.append(middle_row);

    last_row = ([1] * (sideLength + 1) ) + [2];
    extended_board.append(last_row);

    return extended_board;
 
def printBoard(board):
    for row in board:
        print "  ".join(map(str,row));

#print "board:";
#printBoard(Board_foo);
#
#print "extended:";
#printBoard(extend(Board_foo));

def Shuffle( deck ):
    """ Applies the Fisher-Yates shuffle to a list deck.
    This generates uniformly random permutation of deck and returns the output."""
   
    size = len(deck);
    output = copy.deepcopy(deck);

    for i in range(0, size - 1):
        swap = random.randint(0, size - 1);
        output[i], output[swap] = output[swap], output[i];

    return output;

def onBoard(v,n):
    """ Given a vector, determine if it is on the board. """
    if ((-1 <= v[0] <= n + 1) and (0 <= v[1] <= n + 1)): 
        return True;
    else:
        return False;
    

def Gale(board):
    """ Given a filled board, return the winner. """
    xBoard = extend(board);
    sideLength = len(board[0]);

    #printBoard(xBoard);

    # intialize the four-vertex configuration marking the edge between v2 and v3

    # the four vertices in Gale's Algorithm
    #     v1
    #    /  |
    #   /   |
    #  v2--v3
    #  |   /
    #  |  /
    #   v4

    v1 = numpy.array([-1,1]);
    v2 = numpy.array([0,0]);
    v3 = numpy.array([0,1]);
    v4 = numpy.array([1,0]);

    while (onBoard(v4, sideLength)):
        v1_new = numpy.array;
        v2_new = numpy.array;
        v3_new = numpy.array;
        v4_new = numpy.array;

        #print("boop! we are about to move.");
        #print("\t v1 = {0}={1}".format(str(v1), xBoard[v1[0]][v1[1]]));
        #print("\t v2 = {0}={1}".format(str(v2), xBoard[v2[0]][v2[1]]));
        #print("\t v3 = {0}={1}".format(str(v3), xBoard[v3[0]][v3[1]]));
        #print("\t v4 = {0}={1}".format(str(v4), xBoard[v4[0]][v4[1]]));
        
        # PRINT the edge just passed
        #print("\t {0}-{1}".format(str(v2), str(v3)))

        if xBoard[v4[0]][v4[1]] == xBoard[v3[0]][v3[1]]:
            # go right 
            v1_new = v3;
            v2_new = v2;
            v3_new = v4;
            v4_new = v2 + (v2 - v1);
        elif xBoard[v4[0]][v4[1]] == xBoard[v2[0]][v2[1]]:
            # go right
            v1_new = v2;
            v2_new = v4;
            v3_new = v3;
            v4_new = v3 + (v3 - v1);
        
        v1,v2,v3,v4=v1_new,v2_new,v3_new,v4_new;
   
    # if v4 exits the bottom left corner, then `1'st won
    if v4[1] == -1:
        return 1;
    # if v4 exits the top right corner, then `2'nd won
    elif v4[0] == 0:
        return 2;
    else:
       raise ValueError('The Gale algorithm broke: not [* -1] or [0 *].')

def HexRandomBoard(sidelength):
    """ Produces a board by playing out a random sequence of moves.
    This will always generate legal end state of the Hex."""

    moves = [ [i,j] for i in range(0, sidelength) for j in range(0, sidelength)];
    moves = Shuffle(moves);
   
    # initialize an empty board
    board = [ [ 0 for i in range(0, sidelength)] for j in range(0, sidelength)]

    # 1st player plays first
    turn=1;

    while moves:
        move = moves.pop();
        board[move[0]][move[1]] = turn;
        
        # switch players. 1 -> 2 and 2 -> 1
        turn  = (-1)*turn + 3; 

    return board;

def BernoulliRandomBoard(sidelength):
    """ Produces a board by tossing a {R,B} coin for each cell. """

    board = [];
    
    for i in range(0, sidelength):
        row = []
        for j in range(0, sidelength):
            row.append(random.choice([1,2]));
        board.append(row);
        
    return(board);

### Simulator ####

N = args.trials;

halfBoardSize = args.size;
BoardSize = 2*halfBoardSize + 1
criticalCount = numpy.zeros((BoardSize, BoardSize));

#print "Bernoulli Random";
#print BernoulliRandomBoard(BoardSize);

#print "Hex Random";
#print HexRandomBoard(BoardSize);

print "Simulating {0}x{0} with N={1} trials using {2} sampling".format(BoardSize,N, args.sampling);

for x in range(0,BoardSize):
    for y in range(0, BoardSize):
        # the board is centrally symmetric
        # we only need the values above the diagonal
	if (x + y < BoardSize): 
		t = time.time();
		for i in range(1,N):

                   ## Hex final position trials.
                   ## Pick a position where 1 played at [x][y]
                   ## Determine who won in that position.
                   ## Switch the colour of that stone to 2.
                
                   if args.sampling == 'bernoulli':
			board = BernoulliRandomBoard(BoardSize);

		   if args.sampling == 'hex':
                       board = HexRandomBoard(BoardSize);

                   state1 = Gale(board);

                   board[x][y] = 2;
                   state2 = Gale(board);
                   
                   if (state1 != state2):
                       criticalCount[y][x] += 1;   

		print "Criticality[{0}][{1}] = {2} at t={3}".format(x,y,str(criticalCount[y][x]), time.time() - t);

print criticalCount;
